# tests.syncdaemon.test_event_logging - test logging ZG events
#
# Author: Alejandro J. Cura <alecu@canonical.com>
#
# Copyright 2010 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Test the event logging from SyncDaemon into Zeitgeist."""

import os
import uuid

from twisted.internet import defer
from zeitgeist.datamodel import Interpretation, Manifestation

from contrib.testing.testcase import FakeMainTestCase
from ubuntuone.storageprotocol.sharersp import NotifyShareHolder
from ubuntuone.syncdaemon.event_logging import (
        zglog, ZeitgeistListener, ACTOR_UBUNTUONE,
        EVENT_INTERPRETATION_U1_FOLDER_SHARED,
        EVENT_INTERPRETATION_U1_FOLDER_UNSHARED,
        EVENT_INTERPRETATION_U1_SHARE_ACCEPTED,
        EVENT_INTERPRETATION_U1_SHARE_UNACCEPTED,
        MANIFESTATION_U1_CONTACT_DATA_OBJECT, DIRECTORY_MIMETYPE,
        INTERPRETATION_U1_CONTACT, URI_PROTOCOL_U1, STORAGE_NETWORK)
from ubuntuone.syncdaemon.volume_manager import Share, Shared


class MockLogger(object):
    """A mock logger that stores whatever is logged into it."""

    def __init__(self):
        """Initialize this instance."""
        self.events = []

    def log(self, event):
        """Log the event."""
        self.events.append(event)


class ZeitgeistListenerTestCase(FakeMainTestCase):
    """Tests for ZeitgeistListener."""

    def setUp(self):
        """Initialize this instance."""
        super(ZeitgeistListenerTestCase, self).setUp()
        self.patch(zglog, "ZeitgeistLogger", MockLogger)
        self.listener = ZeitgeistListener(self.fs, self.vm)
        self.event_q.subscribe(self.listener)

    def _listen_for(self, event, callback, count=1, collect=False):
        """Setup a EQ listener for the specified event."""
        event_q = self.main.event_q
        class Listener(object):
            """A basic listener to handle the pushed event."""

            def __init__(self):
                self.hits = 0
                self.events = []

            def _handle_event(self, *args, **kwargs):
                self.hits += 1
                if collect:
                    self.events.append((args, kwargs))
                if self.hits == count:
                    event_q.unsubscribe(self)
                    if collect:
                        callback(self.events)
                    elif kwargs:
                        callback((args, kwargs))
                    else:
                        callback(args)

        listener = Listener()
        setattr(listener, 'handle_'+event, listener._handle_event)
        event_q.subscribe(listener)
        return listener

    def test_share_created_with_username_is_logged(self):
        """A ShareCreated event is logged."""
        fake_username = "fake user"
        path = os.path.join(self.vm.root.path, 'shared_path')
        sample_node_id = "node id"
        self.main.fs.create(path, "")
        self.main.fs.set_node_id(path, sample_node_id)

        def fake_create_share(node_id, user, name, access_level, marker):
            """Fake the creation of the share on the server."""
            self.assertIn(marker, self.vm.marker_share_map)
            share_id = self.fs.get_by_mdid(marker).share_id
            self.main.event_q.push('AQ_CREATE_SHARE_OK',
                                   share_id=share_id,
                                   marker=marker)

        self.patch(self.main.action_q, "create_share", fake_create_share)
        self.vm.create_share(path, fake_username, 'shared_name', 'View')

        self.assert_folder_shared_is_logged(path, fake_username)

    def test_share_created_with_email_is_logged(self):
        """A ShareCreated event is logged."""
        fake_username = "fakeuser@somewhere.com"
        path = os.path.join(self.vm.root.path, 'shared_path')
        sample_node_id = "node id"
        self.main.fs.create(path, "")
        self.main.fs.set_node_id(path, sample_node_id)

        def fake_create_share(node_id, user, name, access_level, marker):
            """Fake the creation of the share on the server."""
            self.assertIn(marker, self.vm.marker_share_map)
            self.main.event_q.push('AQ_SHARE_INVITATION_SENT',
                                   marker=marker)

        self.patch(self.main.action_q, "create_share", fake_create_share)
        self.vm.create_share(path, fake_username, 'shared_name', 'View')

        self.assert_folder_shared_is_logged(path, fake_username)

    def assert_folder_shared_is_logged(self, path, fake_username):
        """Assert that the FolderShared event was logged."""

        self.assertEquals(len(self.listener.zg.events), 1)
        event = self.listener.zg.events[0]

        self.assertEquals(event.interpretation,
                          EVENT_INTERPRETATION_U1_FOLDER_SHARED)
        self.assertEquals(event.manifestation,
                          Manifestation.USER_ACTIVITY)
        self.assertEquals(event.actor, ACTOR_UBUNTUONE)

        folder = event.subjects[0]
        self.assertTrue(folder.uri.startswith(URI_PROTOCOL_U1))
        self.assertEquals(folder.interpretation, Interpretation.FOLDER)
        self.assertEquals(folder.manifestation,
                          Manifestation.REMOTE_DATA_OBJECT)
        self.assertTrue(folder.origin.endswith(path))
        self.assertEquals(folder.mimetype, DIRECTORY_MIMETYPE)
        self.assertEquals(folder.storage, STORAGE_NETWORK)

        other_user = event.subjects[1]
        self.assertEquals(other_user.uri, "mailto:" + fake_username)
        self.assertEquals(other_user.interpretation, INTERPRETATION_U1_CONTACT)
        self.assertEquals(other_user.manifestation,
                          MANIFESTATION_U1_CONTACT_DATA_OBJECT)


    @defer.inlineCallbacks
    def test_share_deleted_is_logged(self):
        """Test VolumeManager.delete_share."""
        sample_share_id = "share id"
        sample_node_id = "node id"
        fake_username = "fake user"
        path = os.path.join(self.vm.root.path, 'shared_path')
        self.main.fs.create(path, "")
        self.main.fs.set_node_id(path, sample_node_id)
        share = Shared(path=path, volume_id=sample_share_id,
                       node_id=sample_node_id, other_username=fake_username)
        self.vm.add_shared(share)

        def fake_delete_share(share_id):
            """Fake delete_share."""
            self.assertEqual(share_id, share.volume_id)
            self.main.event_q.push('AQ_DELETE_SHARE_OK', share_id)

        self.patch(self.main.action_q, 'delete_share', fake_delete_share)
        d = defer.Deferred()
        self._listen_for('VM_SHARE_DELETED', d.callback, 1, collect=True)
        self.vm.delete_share(share.volume_id)
        yield d

        self.assertEquals(len(self.listener.zg.events), 1)
        event = self.listener.zg.events[0]

        self.assertEquals(event.interpretation,
                          EVENT_INTERPRETATION_U1_FOLDER_UNSHARED)
        self.assertEquals(event.manifestation,
                          Manifestation.USER_ACTIVITY)
        self.assertEquals(event.actor, ACTOR_UBUNTUONE)

        folder = event.subjects[0]
        self.assertTrue(folder.uri.startswith(URI_PROTOCOL_U1))
        self.assertEquals(folder.interpretation, Interpretation.FOLDER)
        self.assertEquals(folder.manifestation,
                          Manifestation.REMOTE_DATA_OBJECT)
        self.assertTrue(folder.origin.endswith(path))
        self.assertEquals(folder.mimetype, DIRECTORY_MIMETYPE)
        self.assertEquals(folder.storage, STORAGE_NETWORK)

        other_user = event.subjects[1]
        self.assertEquals(other_user.uri, "mailto:" + fake_username)
        self.assertEquals(other_user.interpretation, INTERPRETATION_U1_CONTACT)
        self.assertEquals(other_user.manifestation,
                          MANIFESTATION_U1_CONTACT_DATA_OBJECT)

    def test_share_accepted_is_logged(self):
        """Test that an accepted share event is logged."""
        # initialize the the root
        self.vm._got_root('root_uuid')
        fake_username = "fake user"
        path = os.path.join(self.vm.root.path, 'shared_path')
        self.main.fs.create(path, "")
        share_path = os.path.join(self.shares_dir, 'fake_share')
        share = Share(path=share_path, volume_id='volume_id', node_id="node_id",
                      other_username=fake_username)
        self.vm.add_share(share)

        self.assertEquals(len(self.listener.zg.events), 1)
        event = self.listener.zg.events[0]

        self.assertEquals(event.interpretation,
                          EVENT_INTERPRETATION_U1_SHARE_ACCEPTED)
        self.assertEquals(event.manifestation,
                          Manifestation.USER_ACTIVITY)
        self.assertEquals(event.actor, ACTOR_UBUNTUONE)

        folder = event.subjects[0]
        self.assertTrue(folder.uri.startswith(URI_PROTOCOL_U1))
        self.assertEquals(folder.interpretation, Interpretation.FOLDER)
        self.assertEquals(folder.manifestation,
                          Manifestation.REMOTE_DATA_OBJECT)
        self.assertTrue(folder.origin.endswith(share_path))
        self.assertEquals(folder.mimetype, DIRECTORY_MIMETYPE)
        self.assertEquals(folder.storage, STORAGE_NETWORK)

        other_user = event.subjects[1]
        self.assertEquals(other_user.uri, "mailto:" + fake_username)
        self.assertEquals(other_user.interpretation, INTERPRETATION_U1_CONTACT)
        self.assertEquals(other_user.manifestation,
                          MANIFESTATION_U1_CONTACT_DATA_OBJECT)


    @defer.inlineCallbacks
    def test_share_unaccepted_is_logged(self):
        """Test that an unaccepted share event is logged."""
        fake_username = "fake user"
        d = defer.Deferred()

        share_path = os.path.join(self.main.shares_dir, 'share')
        holder = NotifyShareHolder.from_params(uuid.uuid4(),
                                                     uuid.uuid4(),
                                                     u'fake_share',
                                                     fake_username,
                                                     u'visible_name', 'Read')

        self.main.vm.add_share(Share.from_notify_holder(holder, share_path))
        self._listen_for('VM_VOLUME_DELETED', d.callback, 1, collect=True)
        self.main.event_q.push('SV_SHARE_DELETED', holder.share_id)
        yield d

        self.assertEquals(len(self.listener.zg.events), 2)
        event = self.listener.zg.events[1]

        self.assertEquals(event.interpretation,
                          EVENT_INTERPRETATION_U1_SHARE_UNACCEPTED)
        self.assertEquals(event.manifestation,
                          Manifestation.USER_ACTIVITY)
        self.assertEquals(event.actor, ACTOR_UBUNTUONE)

        folder = event.subjects[0]
        self.assertTrue(folder.uri.startswith(URI_PROTOCOL_U1))
        self.assertEquals(folder.interpretation, Interpretation.FOLDER)
        self.assertEquals(folder.manifestation,
                          Manifestation.REMOTE_DATA_OBJECT)
        self.assertTrue(folder.origin.endswith(share_path))
        self.assertEquals(folder.mimetype, DIRECTORY_MIMETYPE)
        self.assertEquals(folder.storage, STORAGE_NETWORK)

        other_user = event.subjects[1]
        self.assertEquals(other_user.uri, "mailto:" + fake_username)
        self.assertEquals(other_user.interpretation, INTERPRETATION_U1_CONTACT)
        self.assertEquals(other_user.manifestation,
                          MANIFESTATION_U1_CONTACT_DATA_OBJECT)
