# tests.syncdaemon.test_tritcask - tritcask tests
#
# Author: Guillermo Gonzalez <guillermo.gonzalez@canonical.com>
#
# Copyright 2010 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for Tritcask and helper functions."""


import contextlib
import mmap
import os
import time
import uuid
import zlib

from operator import attrgetter

from contrib.testing.testcase import BaseTwistedTestCase
from ubuntuone.syncdaemon.tritcask import (
    TOMBSTONE,
    LIVE,
    HINT,
    INACTIVE,
    DEAD,
    BadCrc,
    DataFile,
    TempDataFile,
    HintFile,
    Tritcask,
    header_size,
    crc32_size,
    crc32_struct,
    header_struct,
    hint_header_size,
    hint_header_struct,
    _get_file_id,
    is_live,
    is_immutable,
    is_hint,
)


class BaseTestCase(BaseTwistedTestCase):
    """Base testcase for Tritcask."""

    def setUp(self):
        BaseTwistedTestCase.setUp(self)
        self.base_dir = self.mktemp('data_dir')
        self.root_dir = self.mktemp('root_dir')

    def build_data(self):
        """Build a random key/value."""
        key = str(uuid.uuid4())
        data = os.urandom(50)
        return key, data


class DataFileTest(BaseTestCase):

    file_class = DataFile

    def test__get_next_file_id(self):
        curr = int(time.time() * 100000)
        self.assertTrue(curr <= int(self.file_class._get_next_file_id()))
        curr = int(time.time() * 100000)
        self.assertTrue(curr <= int(self.file_class._get_next_file_id()))

    def test_exists(self):
        """Tests for exists method."""
        new_file = self.file_class(self.base_dir)
        self.assertTrue(new_file.exists())
        self.assertTrue(os.path.exists(new_file.filename))

    def test_size(self):
        """Test the size property."""
        new_file = self.file_class(self.base_dir)
        self.assertEqual(0, new_file.size)
        new_file.fd.write('foo')
        new_file.fd.flush()
        self.assertEqual(len('foo'), new_file.size)

    def test__open(self):
        """Test the _open private method."""
        new_file = self.file_class(self.base_dir)
        # check that the file is opened
        new_file.fd.write('foo')
        new_file.close()

    def test_close(self):
        """Test the close method."""
        new_file = self.file_class(self.base_dir)
        new_file.close()
        self.assertEqual(None, new_file.fd)

    def test_make_immutable(self):
        """Test for make_immutable method."""
        new_file = self.file_class(self.base_dir)
        new_file.fd.write('foo')
        immutable_file = new_file.make_immutable()
        # the DataFile should be closed
        self.assertEqual(None, new_file.fd)
        self.assertEqual(immutable_file.file_id, new_file.file_id)

    def test_iter_entries(self):
        """Test for iter_entries"""
        db = Tritcask(self.base_dir)
        for i in range(10):
            db.put(i, 'foo%d' % (i,), 'bar%s' % (i,))
        for i, entry in enumerate(db.live_file.iter_entries()):
            self.assertEqual(entry[4], i)
            self.assertEqual(entry[5], 'foo%d' % (i,))
            self.assertEqual(entry[6], 'bar%d' % (i,))
        db.shutdown()

    def test__getitem__(self):
        """Test our slicing support."""
        data_file = self.file_class(self.base_dir)
        tstamp, value_pos, value_sz = data_file.write(0, 'foo', 'bar')
        tstamp1, value1_pos, value1_sz = data_file.write(1, 'foo1', 'bar1')
        self.assertEqual('bar', data_file[value_pos:value_pos+value_sz])
        self.assertEqual('bar1', data_file[value1_pos:value1_pos+value1_sz])

    def test__getitem__no_slice(self):
        """Test that we *only* support slicing."""
        data_file = self.file_class(self.base_dir)
        tstamp, value_pos, value_sz = data_file.write(0, 'foo', 'bar')
        self.assertRaises(ValueError, data_file.__getitem__, value_pos)

    def test_write(self):
        """Test for write method."""
        data_file = self.file_class(self.base_dir)
        tstamp_1, value_1_pos, value_1_sz = data_file.write(0, 'foo', 'bar')
        self.assertEqual(len('bar'), value_1_sz)
        tstamp_2, value_2_pos, value_2_sz = data_file.write(1, 'foo1', 'bar1')
        self.assertEqual(len('bar1'), value_2_sz)
        data_file.close()
        # check that the entry is in the file
        with open(data_file.filename, 'rb') as f:
            raw_data_len = (len('foo') + len('bar') + crc32_size + header_size)
            raw_data = f.read(raw_data_len)
            crc32 = crc32_struct.unpack(raw_data[:crc32_size])[0]
            raw_header = raw_data[crc32_size:crc32_size+header_size]
            header = header_struct.unpack(raw_header)
            tstamp, key_sz, value_sz, row_type = header
            value = raw_data[crc32_size+header_size+key_sz:]
            self.assertEqual(zlib.crc32(raw_header+'foo'+'bar'), crc32)
            self.assertEqual(len('foo'), key_sz)
            self.assertEqual(len('bar'), value_sz)
            self.assertEqual('bar', value)
            self.assertEqual(0, row_type)

    def test_read(self):
        """Test for read method."""
        data_file = self.file_class(self.base_dir)
        orig_tstamp, _, _ = data_file.write(0, 'foo', 'bar')
        tstamp1, _, _ = data_file.write(1, 'foo1', 'bar1')
        fmap = mmap.mmap(data_file.fd.fileno(), 0, access=mmap.ACCESS_READ)
        with contextlib.closing(fmap):
            current_pos = 0
            (crc32, tstamp, key_sz, value_sz, row_type,
                 key, value, pos), new_pos = data_file.read(fmap, current_pos)
            current_pos = new_pos
            self.assertEqual(crc32_size+header_size+key_sz+value_sz, new_pos)
            self.assertEqual(orig_tstamp, tstamp)
            self.assertEqual(len('foo'), key_sz)
            self.assertEqual(len('bar'), value_sz)
            self.assertEqual('bar', value)
            self.assertEqual(0, row_type)
            (crc32, tstamp, key_sz, value_sz, row_type,
                 key, value, pos), new_pos = data_file.read(fmap, current_pos)
            self.assertEqual(
                crc32_size+header_size+key_sz+value_sz+current_pos, new_pos)
            self.assertEqual(tstamp1, tstamp)
            self.assertEqual(len('foo1'), key_sz)
            self.assertEqual(len('bar1'), value_sz)
            self.assertEqual('bar1', value)
            self.assertEqual(1, row_type)
        data_file.close()

    def test_read_bad_crc(self):
        """Test for read method with a bad crc error."""
        data_file = self.file_class(self.base_dir)
        orig_tstamp, _, _ = data_file.write(0, 'foo', 'bar')
        # mess with the data on disk to make this entry crc32 invalid
        # seek to the end of crc32+header+key
        data_file.close()
        with open(data_file.filename, 'r+b') as fd:
            fd.seek(crc32_size+header_size+len('foo'))
            # write a different value -> random bytes
            fd.write(os.urandom(len('bar')))
            fd.flush()
            fmap = mmap.mmap(fd.fileno(), 0, access=mmap.ACCESS_READ)
            with contextlib.closing(fmap):
                self.assertRaises(BadCrc, data_file.read, fmap, 0)


class TempDataFileTest(DataFileTest):
    """Tests for TempDataFile."""

    file_class = TempDataFile

    def test_tempfile_name(self):
        """Test the name of the tempfile isn't LIVE."""
        new_file = self.file_class(self.base_dir)
        self.assertNotIn(LIVE, new_file.filename)


class ImmutableDataFileTest(DataFileTest):

    def test_make_immutable(self):
        """Test for make_immutable."""
        new_file = DataFile(self.base_dir)
        # write some data
        new_file.fd.write('foo')
        immutable_file = new_file.make_immutable()
        # it's the same instance?
        self.assertEqual(immutable_file.make_immutable(), immutable_file)

    def test_make_zombie(self):
        """Test for the make_zombie method."""
        new_file = DataFile(self.base_dir)
        # write some data
        new_file.fd.write('foo')
        immutable_file = new_file.make_immutable()
        # create a zombie
        zombie = immutable_file.make_zombie()
        # it's the same instance?
        self.assertEqual(zombie, immutable_file)
        self.assertFalse(zombie.fd.closed)

    def test_write(self):
        """Test the write fails on immutable files."""
        new_file = DataFile(self.base_dir)
        # write some data
        new_file.fd.write('foo')
        immutable_file = new_file.make_immutable()
        self.assertRaises(NotImplementedError, immutable_file.write, 0, 'foo', 'bar')

    def test__open(self):
        """Test the _open private method."""
        new_file = DataFile(self.base_dir)
        # write some data
        new_file.fd.write('foo')
        immutable_file = new_file.make_immutable()
        self.assertTrue(immutable_file.fd is not None)
        self.assertTrue(immutable_file.fmmap is not None)
        # check that the file is opened only for read
        self.assertRaises(IOError, immutable_file.fd.write, 'foo')
        immutable_file.close()

    def test_close(self):
        """Test the close method."""
        new_file = DataFile(self.base_dir)
        # write some data
        new_file.fd.write('foo')
        immutable_file = new_file.make_immutable()
        immutable_file.close()
        self.assertEqual(None, immutable_file.fd)
        self.assertEqual(None, immutable_file.fmmap)


    def test_iter_entries(self):
        """Test for iter_entries"""
        db = Tritcask(self.base_dir)
        for i in range(10):
            db.put(i, 'foo%d' % (i,), 'bar%s' % (i,))
        file_id = db.live_file.file_id
        db.rotate()
        db.shutdown()
        db = Tritcask(self.base_dir)
        for i, entry in enumerate(db._immutable[file_id].iter_entries()):
            self.assertEqual(entry[4], i)
            self.assertEqual(entry[5], 'foo%d' % (i,))
            self.assertEqual(entry[6], 'bar%d' % (i,))
        db.shutdown()

    def test__getitem__(self):
        """Test our slicing support."""
        rw_file = DataFile(self.base_dir)
        tstamp, value_pos, value_sz = rw_file.write(0, 'foo', 'bar')
        tstamp1, value1_pos, value1_sz = rw_file.write(1, 'foo1', 'bar1')
        data_file = rw_file.make_immutable()
        self.assertEqual('bar', data_file[value_pos:value_pos+value_sz])
        self.assertEqual('bar1', data_file[value1_pos:value1_pos+value1_sz])

    def test__getitem__no_slice(self):
        """Test that we *only* support slicing."""
        rw_file = DataFile(self.base_dir)
        tstamp, value_pos, value_sz = rw_file.write(0, 'foo', 'bar')
        data_file = rw_file.make_immutable()
        self.assertRaises(ValueError, data_file.__getitem__, value_pos)


class HintFileTest(BaseTestCase):
    """Tests for HintFile class."""

    def test_init(self):
        """Test initialization."""
        path = os.path.join(self.base_dir, 'test_hint')
        hint_file = HintFile(path)
        self.assertEqual('w+b', hint_file.fd.mode)

    def test_init_existing(self):
        """Test initialization."""
        path = os.path.join(self.base_dir, 'test_hint_existing')
        hint_file = HintFile(path)
        hint_file.close()
        hint_file = HintFile(path)
        self.assertEqual('rb', hint_file.fd.mode)

    def test_close(self):
        """Test for the close method."""
        path = os.path.join(self.base_dir, 'test_hint_close')
        hint_file = HintFile(path)
        self.assertEqual('w+b', hint_file.fd.mode)
        fd = hint_file.fd
        hint_file.close()
        self.assertTrue(fd.closed)

    def test_contextmanager(self):
        """Test the context manager protocol."""
        path = os.path.join(self.base_dir, 'test_hint_close')
        hint_file = HintFile(path)
        with hint_file as hf:
            self.assertEqual(hint_file, hf)

    def test_iter_entries(self):
        """Test for iter_entries method."""
        db = Tritcask(self.base_dir)
        # create some stuff.
        for i in range(10):
            db.put(i, 'foo%d' % (i,), 'bar%s' % (i,))
        # make the data file inactive and generate the hint.
        immutable_file = db.live_file.make_immutable()
        db.shutdown()
        Tritcask(self.base_dir).shutdown()
        self.assertTrue(immutable_file.has_hint)
        #check that the hint matches the contents in the DB
        hint_file = immutable_file.get_hint_file()
        db = Tritcask(self.base_dir)
        for hint_entry in hint_file.iter_entries():
            self.assertEqual(6, len(hint_entry))
            self.assertTrue(db.get(hint_entry[2], hint_entry[-1]) is not None)
        db.shutdown()

    def test_write(self):
        """Test the write method."""
        hint_file = HintFile(os.path.join(self.base_dir, 'hint_file'))
        tstamp1 = time.time()
        hint_file.write('foo', tstamp1, len('foo'), 0, len('bar'), 100)
        tstamp2 = time.time()
        hint_file.write('foo1', tstamp2, len('foo1'), 1, len('bar1'), 100)
        hint_file.close()
        # check that the entry is in the file
        with open(hint_file.path, 'rb') as f:
            header = hint_header_struct.unpack(f.read(hint_header_size))
            tstamp, key_sz, row_type, value_sz, value_pos = header
            self.assertEqual(tstamp1, tstamp)
            self.assertEqual(len('foo'), key_sz)
            self.assertEqual(0, row_type)
            self.assertEqual(len('bar'), value_sz)
            self.assertEqual(100, value_pos)
            key = f.read(key_sz)
            self.assertEqual('foo', key)
            # read the second entry
            header = hint_header_struct.unpack(f.read(hint_header_size))
            tstamp, key_sz, row_type, value_sz, value_pos = header
            self.assertEqual(tstamp2, tstamp)
            self.assertEqual(len('foo1'), key_sz)
            self.assertEqual(1, row_type)
            self.assertEqual(len('bar1'), value_sz)
            self.assertEqual(100, value_pos)
            key = f.read(key_sz)
            self.assertEqual('foo1', key)


class LowLevelTest(BaseTestCase):
    """Tests for low level methods and functions.

    Test the on-disk format, writes and reads.
    """

    def setUp(self):
        BaseTestCase.setUp(self)
        self.db = Tritcask(self.base_dir)

    def tearDown(self):
        BaseTestCase.tearDown(self)
        self.db.shutdown()

    def test_get_file_id(self):
        """Test for _get_file_id function."""
        filename = os.path.split(self.db.live_file.filename)[1]
        self.assertEqual(self.db.live_file.file_id, _get_file_id(filename))

    def test_is_hint(self):
        """Test for is_hint function."""
        filename = DataFile._get_next_file_id() + LIVE
        self.assertFalse(is_hint(filename))
        filename = DataFile._get_next_file_id() + HINT
        self.assertTrue(is_hint(filename))

    def test_is_immutable(self):
        """Test for is_immutable function."""
        filename = DataFile._get_next_file_id() + LIVE
        self.assertFalse(is_immutable(filename))
        filename = DataFile._get_next_file_id() + INACTIVE
        self.assertTrue(is_immutable(filename))

    def test_is_live(self):
        """Test for is_live function."""
        filename = DataFile._get_next_file_id() + HINT
        self.assertFalse(is_live(filename))
        filename = DataFile._get_next_file_id() + LIVE
        self.assertTrue(is_live(filename))

    def test_get_value(self):
        """Test _get_value method."""
        self.db.put(0, 'foo', 'bar')
        value = self.db._get_value(self.db.live_file.file_id,
                               crc32_size+header_size+len('foo'), len('bar'))
        self.assertEqual(value, 'bar')

    def test_get_value_different_file_ids(self):
        """Test _get_value with different file_id."""
        self.db.put(0, 'foo', 'bar')
        old_file_id = self.db.live_file.file_id
        # shutdown and rename the file.
        self.db.live_file.make_immutable()
        self.db.shutdown()
        self.db = Tritcask(self.base_dir)
        self.db.put(1, 'foo1', 'bar1')
        self.assertEqual(1, len(self.db._immutable))
        # read from the old file
        value = self.db._get_value(
            old_file_id, crc32_size+header_size+len('foo'), len('bar'))
        self.assertEqual(value, 'bar')
        # read from the current file
        value = self.db._get_value(self.db.live_file.file_id,
                                   crc32_size+header_size+len('foo1'),
                                   len('bar1'))
        self.assertEqual(value, 'bar1')

    def test_shutdown(self):
        """Test shutdown."""
        # create 1 inactive files, with 5 items each
        for i in range(5):
            self.db.put(i, 'foo%d' % (i,), 'bar%s' % (i,))
        # make the data file inactive and generate the hint.
        self.db.live_file.make_immutable()
        self.db.shutdown()
        self.db = Tritcask(self.base_dir)
        for i in range(5, 10):
            self.db.put(i, 'foo%d' % (i,), 'bar%s' % (i,))
        # iterate over all values, in order to open all inactive data
        # files.
        for rtype, key in self.db._keydir.keys():
            self.db.get(rtype, key)
        # check that we have 2 open files
        self.assertEqual(1, len(self.db._immutable))
        self.assertNotEqual(None, self.db.live_file)
        self.db.shutdown()
        self.assertEqual(None, self.db.live_file)
        self.assertEqual(0, len(self.db._immutable))

    def test_rotate_files(self):
        """Test data file rotation."""
        # add a key/value
        self.db.put(0, 'foo', 'bar')
        files = sorted(os.listdir(self.db.base_path))
        self.assertEqual(1, len(files), files)
        self.db.rotate()
        files = sorted(os.listdir(self.db.base_path))
        self.assertEqual(2, len(files), files)
        self.assertIn(INACTIVE, files[0])
        self.assertIn(LIVE, files[1])
        # add a new value to trigger the creation of the new file
        self.db.put(0, 'foo', 'bar1')
        self.assertEqual(2, len(files), files)
        self.assertIn(INACTIVE, files[0])
        self.assertIn(LIVE, files[1])
        self.db.shutdown()

    def test_rotate_empty_live_file(self):
        """Test data file rotation with empty live file."""
        self.db.rotate()
        files = sorted(os.listdir(self.db.base_path))
        self.assertEqual(1, len(files), files)
        self.assertIn(LIVE, files[0])
        self.assertEqual(self.db.live_file.filename,
                         os.path.join(self.db.base_path, files[0]))
        self.db.shutdown()


class InitTest(BaseTestCase):
    """Init tests."""

    def test_initialize_new(self):
        """Simple initialization."""
        os.rmdir(self.base_dir)
        db = Tritcask(self.base_dir)
        db.shutdown()

    def test_initialize_existing(self):
        """Initialize with a existing db."""
        db = Tritcask(self.base_dir)
        db.shutdown()
        db = Tritcask(self.base_dir)
        db.shutdown()

    def test_initialize_bad_path(self):
        """Initialize with a invalid path."""
        path = os.path.join(self.base_dir, 'foo')
        open(path, 'w').close()
        self.assertRaises(ValueError, Tritcask, path)

    def test_find_data_files(self):
        """Test the _find_data_files method."""
        filenames = []
        for i in range(10):
            data_file = DataFile(self.base_dir)
            data_file.write(i, 'foo_%d' % (i,), 'bar_%d' % (i,))
            immutable_file = data_file.make_immutable()
            filenames.append(immutable_file.filename)
            immutable_file.close()
        db = Tritcask(self.base_dir)
        self.assertNotIn(db.live_file.filename, filenames)
        files = [fn.filename for fn in sorted(db._immutable.values(),
                                              key=attrgetter('filename'))]
        self.assertEqual(files, filenames)
        db.shutdown()

    def test_build_keydir_on_init(self):
        """Test _build_keydir method."""
        db = Tritcask(self.base_dir)
        # create some stuff.
        for i in range(10):
            db.put(i, 'foo%d' % (i,), 'bar%s' % (i,))
        db.shutdown()
        old_keydir = db._keydir
        db = Tritcask(self.base_dir)
        # check that the keydir is the same.
        self.assertEqual(old_keydir, db._keydir)
        db.shutdown()

    def test_build_keydir_with_bad_data(self):
        """Test _build_keydir method with a bad data file."""
        db = Tritcask(self.base_dir)
        # create some stuff.
        for i in range(10):
            db.put(i, 'foo%d' % (i,), 'bar%s' % (i,))
        live_filename = db.live_file.filename
        db.shutdown()
        with open(live_filename, 'r+') as fd:
            # seek to the middle of the last bytes
            fd.seek(-5, os.SEEK_END)
            fd.write(os.urandom(len('this is bad data.')))
            fd.flush()
        old_keydir = db._keydir
        db = Tritcask(self.base_dir)
        # check that the keydir is the expected:
        # the new keydir should have entries from 0-8
        # remove the extra entry from the old keydir
        old_keydir.pop((9, 'foo9'))
        self.assertEqual(old_keydir, db._keydir)
        db.shutdown()

    def test_build_keydir_with_hint(self):
        """Test _build_keydir using a hint file."""
        db = Tritcask(self.base_dir)
        # create some stuff.
        for i in range(10):
            db.put(i, 'foo%d' % (i,), 'bar%s' % (i,))
        # make the data file inactive and generate the hint.
        db.live_file.make_immutable()
        old_keydir = db._keydir
        db.shutdown()
        Tritcask(self.base_dir).shutdown()
        # now create a new DB using the hint
        db = Tritcask(self.base_dir)
        self.assertEqual(old_keydir, db._keydir)
        db.shutdown()

    def test_build_keydir_with_multiple_hint(self):
        """Test _build_keydir using a several hint files."""
        db = Tritcask(self.base_dir)
        # create some stuff.
        for i in range(10):
            db.put(i, 'foo%d' % (i,), 'bar%s' % (i,))
        # make the data file inactive and generate the hint.
        db.live_file.make_immutable()
        db.shutdown()
        # create more stuff.
        db = Tritcask(self.base_dir)
        for i in range(20, 30):
            db.put(i, 'foo%d' % (i,), 'bar%s' % (i,))
        # make the data file inactive and generate the hint.
        db.live_file.make_immutable()
        old_keydir = db._keydir
        db.shutdown()
        Tritcask(self.base_dir).shutdown()
        # now create a new DB using the hint
        db = Tritcask(self.base_dir)
        self.assertEqual(old_keydir, db._keydir)
        db.shutdown()

    def test_build_keydir_build_hint(self):
        """Test _build_keydir with build_hint=True."""
        db = Tritcask(self.base_dir)
        # create some stuff.
        for i in range(10):
            db.put(i, 'foo%d' % (i,), 'bar%s' % (i,))
        # make the data file inactive and generate the hint.
        hint_filename = db.live_file.make_immutable().hint_filename
        db.shutdown()
        db = Tritcask(self.base_dir)
        self.assertTrue(os.path.exists(hint_filename))
        #check that the hint matches the contents in the DB
        hint_file = HintFile(hint_filename)
        for hint_entry in hint_file.iter_entries():
            self.assertEqual(6, len(hint_entry))
            self.assertTrue(db.get(hint_entry[2], hint_entry[-1]) is not None)
        db.shutdown()

    def test_build_keydir(self):
        """Test _build_keydir method."""
        db = Tritcask(self.base_dir)
        # create some stuff.
        for i in range(10):
            db.put(i, 'foo%d' % (i,), 'bar%s' % (i,))
        old_keydir = db._keydir
        # shutdown (close the files)
        db.shutdown()
        db = Tritcask(self.base_dir)
        self.assertEqual(old_keydir, db._keydir)
        db.shutdown()

    def test_build_keydir_with_dead_rows(self):
        """Test _build_keydir method with TOMBSTONE rows."""
        db = Tritcask(self.base_dir)
        # create some stuff.
        for i in range(10):
            db.put(i, 'foo%d' % (i,), 'bar%s' % (i,))
        # delete half the nodes
        for i in range(0, 10, 2):
            db.delete(i, 'foo%d' % (i,))
        for i in range(10, 20):
            db.put(i, 'foo%d' % (i,), 'bar%s' % (i,))
        old_keydir = db._keydir
        # shutdown (close the files)
        db.shutdown()
        db = Tritcask(self.base_dir)
        self.assertEqual(old_keydir, db._keydir)
        db.shutdown()


class BasicTest(BaseTestCase):

    def setUp(self):
        BaseTestCase.setUp(self)
        self.db = Tritcask(self.base_dir)

    def tearDown(self):
        BaseTestCase.tearDown(self)
        self.db.shutdown()

    def test_put(self):
        """Basic test for the put method."""
        key, data = self.build_data()
        self.db.put(0, key, data)
        self.assertEqual(1, len(self.db._keydir.keys()))
        # check that the entry is in the file
        with open(self.db.live_file.filename, 'r+b') as f:
            raw_data_len = (len(key) + len(data) + crc32_size + header_size)
            raw_data = f.read(raw_data_len)
            header = header_struct.unpack(
                raw_data[crc32_size:crc32_size+header_size])
            tstamp, key_sz, value_sz, row_type = header
            self.assertEqual(data, raw_data[crc32_size+header_size+key_sz:])
        self.assertEqual(self.db.get(0, key), data)
        # add new entry and check
        key_1, data_1 = self.build_data()
        self.db.put(1, key_1, data_1)
        self.assertEqual(2, len(self.db._keydir.keys()))
        self.assertEqual(self.db.get(0, key), data)
        self.assertEqual(self.db.get(1, key_1), data_1)

    def test_delete(self):
        """Basic test for the delete method."""
        key, data = self.build_data()
        self.db.put(0, key, data)
        self.db.delete(0, key)
        # check that the TOMBSTONE is there for these keys
        with open(self.db.live_file.filename, 'r+b') as f:
            raw_data_len = len(key) + len(TOMBSTONE) + crc32_size + header_size
            f.seek(-1*raw_data_len, os.SEEK_END)
            raw_data = f.read(raw_data_len)
            self.assertEqual(TOMBSTONE,
                             raw_data[crc32_size+header_size+len(key):])

    def test_get(self):
        """Basic test for the get method."""
        key, data = self.build_data()
        self.db.put(0, key, data)
        self.assertEqual(self.db.get(0, key), data)
        self.db.delete(0, key)

    def test_put_None_key(self):
        """test put with key=None."""
        _, data = self.build_data()
        self.assertRaises(ValueError, self.db.put, 0, None, data)

    def test_get_None_key(self):
        """test put with key=None."""
        key, data = self.build_data()
        self.db.put(0, key, data)
        self.assertRaises(ValueError, self.db.get, 0, None)

    def test_delete_None_key(self):
        """test put with key=None."""
        key, data = self.build_data()
        self.assertRaises(ValueError, self.db.delete, 0, None)


class MergeTests(BaseTestCase):

    def _add_data(self, db, size=100):
        """Add random data to a db."""
        key, value = self.build_data()
        for i in range(size):
            db.put(0, key, value)

    def test_simple_merge(self):
        """Test a simple merge."""
        # create 3 immutable files
        db = Tritcask(self.base_dir)
        self._add_data(db, 100)
        db.rotate()
        self._add_data(db, 100)
        db.rotate()
        self._add_data(db, 100)
        db.shutdown()
        db = Tritcask(self.base_dir)
        keydir_to_merge = db._keydir.copy()
        # add data to the live file
        self._add_data(db, 100)
        immutable_fnames = [ifile.filename for ifile in db._immutable.values()]
        # do the merge.
        data_file = db.merge(db._immutable)
        for entry in data_file.iter_entries():
            (crc32, tstamp, key_sz, value_sz, row_type,
                key, value, value_pos) = entry
            keydir_entry = keydir_to_merge[(row_type, key)]
            (old_file_id, old_tstamp, old_value_sz, old_value_pos) = \
                    keydir_entry
            old_value = db._get_value(old_file_id, old_value_pos, old_value_sz)
            self.assertEqual(old_value, value)
        for fname in immutable_fnames:
            self.assertFalse(os.path.exists(fname))
            self.assertTrue(os.path.exists(fname.replace(INACTIVE, DEAD)))


