# ubuntuone.syncdaemon.event_logging - listen for SD events, log into ZG
#
# Author: Alejandro J. Cura <alecu@canonical.com>
#
# Copyright 2010 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Event logging from SyncDaemon into Zeitgeist."""

from zeitgeist.datamodel import Event, Interpretation, Manifestation, Subject

from ubuntuone.eventlog import zglog
from ubuntuone.syncdaemon.volume_manager import Share

ACTOR_UBUNTUONE = "dbus://com.ubuntuone.SyncDaemon.service"
DIRECTORY_MIMETYPE = "inode/directory"
EVENT_INTERPRETATION_U1_FOLDER_SHARED = "u1://FolderShared"
EVENT_INTERPRETATION_U1_FOLDER_UNSHARED = "u1://FolderUnshared"
EVENT_INTERPRETATION_U1_SHARE_ACCEPTED = "u1://ShareAccepted"
EVENT_INTERPRETATION_U1_SHARE_UNACCEPTED = "u1://ShareUnaccepted"
MANIFESTATION_U1_CONTACT_DATA_OBJECT = "u1://ContactDataObject"
INTERPRETATION_U1_CONTACT = "u1://Contact"
URI_PROTOCOL_U1 = "ubuntuone:"
STORAGE_NETWORK = "net"

class ZeitgeistListener(object):
    """An Event Queue listener that logs into ZG."""

    def __init__(self, fsm, vm):
        """Initialize this instance."""
        self.fsm = fsm
        self.vm = vm
        self.zg = zglog.ZeitgeistLogger()

    def handle_AQ_CREATE_SHARE_OK(self, share_id=None, marker=None):
        """Log the 'directory shared thru the server' event."""
        share = self.vm.shared[share_id]
        self.log_folder_shared(share, share_id)

    def handle_AQ_SHARE_INVITATION_SENT(self, marker):
        """Log the 'directory shared thru http' event."""
        share = self.vm.marker_share_map[marker]
        mdo = self.fsm.get_by_mdid(marker)
        self.log_folder_shared(share, mdo.share_id)

    def log_folder_shared(self, share, share_id):
        """Log the 'directory shared' event."""
        fullpath = self.fsm.get_abspath(share_id, share.path)

        folder = Subject.new_for_values(
            uri=URI_PROTOCOL_U1 + str(share.node_id),
            interpretation=Interpretation.FOLDER,
            manifestation=Manifestation.REMOTE_DATA_OBJECT,
            origin="file:///" + fullpath,
            mimetype=DIRECTORY_MIMETYPE,
            storage=STORAGE_NETWORK)

        other_username = share.other_username
        other_user = Subject.new_for_values(
            uri="mailto:" + other_username,
            interpretation=INTERPRETATION_U1_CONTACT,
            manifestation=MANIFESTATION_U1_CONTACT_DATA_OBJECT)

        event = Event.new_for_values(
            interpretation=EVENT_INTERPRETATION_U1_FOLDER_SHARED,
            manifestation=Manifestation.USER_ACTIVITY,
            actor=ACTOR_UBUNTUONE,
            subjects=[folder, other_user])

        self.zg.log(event)

    def handle_VM_SHARE_DELETED(self, share):
        """Log the share deleted event."""
        folder = Subject.new_for_values(
            uri=URI_PROTOCOL_U1 + str(share.node_id),
            interpretation=Interpretation.FOLDER,
            manifestation=Manifestation.REMOTE_DATA_OBJECT,
            origin="file:///" + share.path,
            mimetype=DIRECTORY_MIMETYPE,
            storage=STORAGE_NETWORK)

        other_username = share.other_username
        other_user = Subject.new_for_values(
            uri="mailto:" + other_username,
            interpretation=INTERPRETATION_U1_CONTACT,
            manifestation=MANIFESTATION_U1_CONTACT_DATA_OBJECT)

        event = Event.new_for_values(
            interpretation=EVENT_INTERPRETATION_U1_FOLDER_UNSHARED,
            manifestation=Manifestation.USER_ACTIVITY,
            actor=ACTOR_UBUNTUONE,
            subjects=[folder, other_user])

        self.zg.log(event)

    def handle_VM_SHARE_CREATED(self, share_id):
        """Log the share accepted event."""
        share = self.vm.shares[share_id]

        folder = Subject.new_for_values(
            uri=URI_PROTOCOL_U1 + str(share.node_id),
            interpretation=Interpretation.FOLDER,
            manifestation=Manifestation.REMOTE_DATA_OBJECT,
            origin="file:///" + share.path,
            mimetype=DIRECTORY_MIMETYPE,
            storage=STORAGE_NETWORK)

        other_username = share.other_username
        other_user = Subject.new_for_values(
            uri="mailto:" + other_username,
            interpretation=INTERPRETATION_U1_CONTACT,
            manifestation=MANIFESTATION_U1_CONTACT_DATA_OBJECT)

        event = Event.new_for_values(
            interpretation=EVENT_INTERPRETATION_U1_SHARE_ACCEPTED,
            manifestation=Manifestation.USER_ACTIVITY,
            actor=ACTOR_UBUNTUONE,
            subjects=[folder, other_user])

        self.zg.log(event)

    def log_share_unaccepted(self, share):
        """Log the share unaccepted event."""
        folder = Subject.new_for_values(
            uri=URI_PROTOCOL_U1 + str(share.node_id),
            interpretation=Interpretation.FOLDER,
            manifestation=Manifestation.REMOTE_DATA_OBJECT,
            origin="file:///" + share.path,
            mimetype=DIRECTORY_MIMETYPE,
            storage=STORAGE_NETWORK)

        other_username = share.other_username
        other_user = Subject.new_for_values(
            uri="mailto:" + other_username,
            interpretation=INTERPRETATION_U1_CONTACT,
            manifestation=MANIFESTATION_U1_CONTACT_DATA_OBJECT)

        event = Event.new_for_values(
            interpretation=EVENT_INTERPRETATION_U1_SHARE_UNACCEPTED,
            manifestation=Manifestation.USER_ACTIVITY,
            actor=ACTOR_UBUNTUONE,
            subjects=[folder, other_user])

        self.zg.log(event)

    def handle_VM_VOLUME_DELETED(self, volume):
        """Log the share/UDF unaccepted event."""
        if isinstance(volume, Share):
            self.log_share_unaccepted(volume)
