# ubuntuone.syncdaemon.tritcask - key/value store
#
# Author: Guillermo Gonzalez <guillermo.gonzalez@canonical.com>
#
# Copyright 2010 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""bitcask-like (row_type,key)/value store."""

import os
import struct
import zlib
import uuid
import mmap
import contextlib
import time
import tempfile
import itertools

from operator import attrgetter

crc32_fmt = '>i'
crc32_size = struct.calcsize(crc32_fmt)
crc32_struct = struct.Struct(crc32_fmt)

header_fmt = '>diii'
header_size = struct.calcsize(header_fmt)
header_struct = struct.Struct(header_fmt)

hint_header_fmt = '>diiii'
hint_header_size = struct.calcsize(hint_header_fmt)
hint_header_struct = struct.Struct(hint_header_fmt)

TOMBSTONE = str(uuid.uuid5(uuid.NAMESPACE_OID, 'TOMBSTONE'))

LIVE = '.live'
INACTIVE = '.inactive'
HINT = '.hint'
DEAD = '.dead'

VERSION = 'v1'
FILE_SUFFIX = '.tritcask-%s.data' % VERSION


class BadCrc(Exception):
    """A Exception for Bad CRC32."""


def _get_file_id(filename):
    """Return the file_id for this filename."""
    return filename.split('.')[0]


def is_hint(filename):
    return HINT in filename


def is_live(filename):
    return LIVE in filename


def is_immutable(filename):
    return INACTIVE in filename


class DataFile(object):
    """Class that encapsulates data file handling."""

    def __init__(self, base_path, filename=None):
        """Create a DataFile instance.

        If filename is None a new file is created.
        """
        if filename is None:
            filename = self._get_next_file_id() + LIVE + FILE_SUFFIX
        self.filename = os.path.join(base_path, filename)
        self.file_id = _get_file_id(filename)
        self.hint_filename = os.path.join(base_path, filename + HINT)
        self.fd = None
        self._open()

    @property
    def size(self):
        return os.stat(self.filename).st_size

    @staticmethod
    def _get_next_file_id():
        """Return the next file id."""
        return str(int(time.time() * 100000))

    @property
    def has_hint(self):
        """Returns true if there is a hint file on-disk."""
        return os.path.exists(self.hint_filename)

    def exists(self):
        return os.path.exists(self.filename)

    def make_immutable(self):
        """Make this data file immutable."""
        self.close()
        new_name = self.filename.replace(LIVE, INACTIVE)
        os.rename(self.filename, new_name)
        return ImmutableDataFile(*os.path.split(new_name))

    def _open(self):
        if not os.path.exists(self.filename):
            # create the file and open it for append
            open(self.filename, 'wb').close()
        self.fd = open(self.filename, 'a+b')

    def close(self):
        """Close the file descriptor"""
        if self.fd:
            self.fd.flush()
            os.fsync(self.fd.fileno())
            self.fd.close()
            self.fd = None

    def iter_entries(self):
        """Returns a generator for the entries in the mmaped file."""
        fmmap = mmap.mmap(self.fd.fileno(), 0, access=mmap.ACCESS_READ)
        with contextlib.closing(fmmap):
            current_pos = 0
            while True:
                try:
                    entry, new_pos = self.read(fmmap, current_pos)
                    current_pos = new_pos
                    yield entry
                except (EOFError, BadCrc):
                    raise StopIteration

    def __getitem__(self, item):
        """__getitem__ to support slicing and *only* slicing."""
        if isinstance(item, slice):
            self.fd.seek(item.start)
            return self.fd.read(item.stop-item.start)
        else:
            raise ValueError('Only slice is supported')

    def write(self, row_type, key, value):
        """Write an entry to the file."""
        key_sz = len(key)
        value_sz = len(value)
        tstamp = time.time()
        header = header_struct.pack(tstamp, key_sz, value_sz, row_type)
        crc32 = crc32_struct.pack(zlib.crc32(header+key+value))
        self.fd.write(crc32+header)
        self.fd.write(key)
        value_pos = self.fd.tell()
        self.fd.write(value)
        self.fd.flush()
        return tstamp, value_pos, value_sz

    def read(self, fmmap, current_pos):
        """Read a single entry from the current position."""
        crc32_bytes = fmmap[current_pos:current_pos+crc32_size]
        current_pos += crc32_size
        header = fmmap[current_pos:current_pos+header_size]
        current_pos += header_size
        if header == '' or crc32_bytes == '':
            # reached EOF
            raise EOFError
        crc32 = crc32_struct.unpack(crc32_bytes)[0]
        tstamp, key_sz, value_sz, row_type = header_struct.unpack(header)
        key = fmmap[current_pos:current_pos+key_sz]
        current_pos += key_sz
        value_pos = current_pos
        value = fmmap[current_pos:current_pos+value_sz]
        current_pos += value_sz
        # verify the crc32 of the data
        if zlib.crc32(header+key+value) == crc32:
            return (crc32, tstamp, key_sz,
                    value_sz, row_type, key, value, value_pos), current_pos
        else:
            raise BadCrc(crc32, zlib.crc32(header+key+value))

    def get_hint_file(self):
        """Open and return the hint file."""
        return HintFile(self.hint_filename)


class ImmutableDataFile(DataFile):
    """A immutable data file."""

    def __init__(self, base_path, filename):
        """Create a ImmutableDataFile instance"""
        self.fmmap = None
        super(ImmutableDataFile, self).__init__(base_path, filename)

    def make_immutable(self):
        """A no-op."""
        return self

    def write(self, *args):
        """Raise a NotImplementedError"""
        raise NotImplementedError

    def make_zombie(self):
        """Rename the file but leave it open.

        Actually close, rename and open it again.
        """
        new_name = self.filename.replace(INACTIVE, DEAD)
        self.close()
        os.rename(self.filename, new_name)
        self.filename = new_name
        self._open()
        return self

    def _open(self):
        self.fd = open(self.filename, 'rb')
        fmmap = mmap.mmap(self.fd.fileno(), 0, access=mmap.ACCESS_READ)
        self.fmmap = fmmap

    def close(self):
        """Close the file descriptor and mmap."""
        if self.fmmap:
            self.fmmap.close()
            self.fmmap = None
        super(ImmutableDataFile, self).close()

    def iter_entries(self):
        """Returns a generator for the entries in the mmaped file."""
        current_pos = 0
        while True:
            try:
                entry, new_pos = self.read(self.fmmap, current_pos)
                current_pos = new_pos
                yield entry
            except (EOFError, BadCrc):
                raise StopIteration

    def __getitem__(self, item):
        """__getitem__ to support slicing and *only* slicing."""
        if isinstance(item, slice):
            return self.fmmap[item]
        else:
            raise ValueError('Only slice is supported')


class TempDataFile(DataFile):

    def __init__(self, base_path):
        """Create a TempDataFile instance."""
        self.temp_name = '.' + tempfile.mktemp(dir='')
        fname = self._get_next_file_id() + self.temp_name + FILE_SUFFIX
        super(TempDataFile, self).__init__(base_path, filename=fname)

    def make_immutable(self):
        """Make this data file immutable."""
        self.close()
        new_name = self.filename.replace(self.temp_name, INACTIVE)
        os.rename(self.filename, new_name)
        return ImmutableDataFile(*os.path.split(new_name))


class HintFile(object):
    """A hint file."""

    def __init__(self, path):
        """Create the instance."""
        self.path = path
        self.tempfile = None
        if os.path.exists(self.path):
            # if it's there open only for read
            self.fd = open(self.path, 'rb')
        else:
            # this is a new hint file, lets create it as a tempfile.
            self.tempfile = tempfile.mktemp(dir=os.path.dirname(self.path))
            self.fd = open(self.tempfile, 'w+b')

    def iter_entries(self):
        """Return a generator over the hint entries."""
        fmap = mmap.mmap(self.fd.fileno(), 0, access=mmap.ACCESS_READ)
        with contextlib.closing(fmap):
            current_pos = 0
            while True:
                header = fmap[current_pos:current_pos+hint_header_size]
                current_pos += hint_header_size
                if header == '':
                    raise StopIteration
                tstamp, key_sz, row_type, value_sz, value_pos = \
                        hint_header_struct.unpack(header)
                key = fmap[current_pos:current_pos+key_sz]
                current_pos += key_sz
                yield (tstamp, key_sz, row_type, value_sz, value_pos, key)

    def __enter__(self):
        """Do nothing, the fd should be already opened."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Close the fd and flush evetything to disk.

        Also rename the tempfile to the hint filename.
        """
        # make sure the hint data is on disk
        self.close()

    def write(self, key, tstamp, key_sz, row_type, value_sz, value_pos):
        """Write a hint entry to the file."""
        self.fd.write(hint_header_struct.pack(
            tstamp, key_sz, row_type, value_sz, value_pos))
        self.fd.write(key)

    def close(self):
        """Close the fd."""
        if self.fd:
            self.fd.flush()
            os.fsync(self.fd.fileno())
            self.fd.close()
        # if this is a new hint file, rename it to the real path.
        if self.tempfile:
            os.rename(self.tempfile, self.path)


class Tritcask(object):
    """Implementation of a bitcask-like (row_type,key)/value store.

    The key-value pairs are stored in a "append-only" file, but we also have a
    namespace/row_type.
    Each record has the following format:

    ---------------------------------------------------------------
    |                     header                    |    bytes    |
    ---------------------------------------------------------------
    | crc32 | tstamp | key_sz | value_sz | row_type | key | value |
    ---------------------------------------------------------------
    |   4   |   4    |   4    |    4     |    4     |  X  |   X   |
    ---------------------------------------------------------------

    The on-disk layout is composed of live and inactive files (+ hint files),
    there can be only one live data file and many inactive (immutable) data
    files. Each of the immutable data files might have a hint file to speedup
    the startup, the format of the hint file is:

    -------------------------------------------------------------
    |                     header                        | bytes |
    -------------------------------------------------------------
    | tstamp | key_sz | row_type | value_sz | value_pos |  key  |
    -------------------------------------------------------------
    |    4   |    4   |     4    |    4     |    4      |   X   |
    -------------------------------------------------------------

    The inactive files are compacted/merged into a smaller (also inactive)
    data files in order to keep fragmentation under control.
    """

    def __init__(self, path):
        """Initialize the instance."""
        self._keydir = {}
        self.base_path = path
        if not os.path.exists(self.base_path):
            os.makedirs(self.base_path)
        elif not os.path.isdir(self.base_path):
            raise ValueError('path must be a directory.')
        self.live_file = None
        self._immutable = {}
        self._find_data_files()
        self._build_keydir()

    def shutdown(self):
        """Shutdown and close all open files."""
        if self.live_file:
            self.live_file.close()
            self.live_file = None
        for data_file in self._immutable.values():
            data_file.close()
        self._immutable.clear()

    def _find_data_files(self):
        """Collect the files we need to work with."""
        files = os.listdir(self.base_path)
        for filename in files:
            if is_hint(filename):
                continue
            if is_live(filename):
                self.live_file = DataFile(self.base_path, filename)
            elif is_immutable(filename):
                # it's a immutable file
                data_file = ImmutableDataFile(self.base_path, filename)
                self._immutable[data_file.file_id] = data_file
        # check if we found a live data file
        # if not, define one (it will be created later)
        if self.live_file is None:
            # it's a clent start, let's create the first file
            self.live_file = DataFile(self.base_path)

    def rotate(self):
        # if the file exists and the size is 0, don't rotate it.
        if self.live_file.exists() and self.live_file.size == 0:
            return
        # add the current file to the "immutable" list
        self.live_file.close()
        data_file = self.live_file.make_immutable()
        self._immutable[data_file.file_id] = data_file
        # create a new live data file
        self.live_file = DataFile(self.base_path)

    def _build_keydir(self):
        """Build the keydir."""
        # sort the files by name, in order to load from older -> newest
        for data_file in sorted(self._immutable.values(),
                                        key=attrgetter('filename')):
            if data_file.has_hint:
                # load directly from the hint file.
                self._load_from_hint(data_file)
            elif data_file.exists() and data_file.size > 0:
                self._load_from_data(data_file)
        if self.live_file.exists() and self.live_file.size > 0:
            self._load_from_data(self.live_file)

    def _load_from_hint(self, data_file):
        """Load keydir contents from a hint file."""
        hint_file = data_file.get_hint_file()
        for hint_entry in hint_file.iter_entries():
            tstamp, key_sz, row_type, value_sz, value_pos, key = hint_entry
            idx_value = (data_file.file_id, tstamp, value_sz, value_pos)
            self._keydir[(row_type, key)] = idx_value

    def _load_from_data(self, data_file):
        """Load keydir info from a data file.

        if hint_file != None build the hint file for this data file.
        """
        build_hint = not data_file.has_hint
        hint_idx = {}
        for entry in data_file.iter_entries():
            (crc32, tstamp, key_sz, value_sz, row_type,
                key, value, value_pos) = entry
            # only add it to the index if value isn't TOMBSTONE
            if value == TOMBSTONE:
                # the record is dead, check if need to remove it from
                # the indexes
                self._keydir.pop((row_type, key), None)
                hint_idx.pop(key, None)
            else:
                idx_value = (data_file.file_id, tstamp, value_sz, value_pos)
                self._keydir[(row_type, key)] = idx_value
                if build_hint:
                    hint_entry = (tstamp, key_sz, row_type, value_sz, value_pos)
                    hint_idx[key] = hint_entry
        if build_hint:
            with data_file.get_hint_file() as hint_file:
                for key, value in hint_idx.iteritems():
                    hint_file.write(key, *value)

    def _get_value(self, file_id, value_pos, value_sz):
        """Get the value for file_id, value_pos."""
        if file_id == self.live_file.file_id:
            # it's the current live file
            return self.live_file[value_pos:value_pos+value_sz]
        else:
            return self._immutable[file_id][value_pos:value_pos+value_sz]

    def put(self, row_type, key, value):
        """Put key/value in the store."""
        # now build the real record
        if key is None:
            raise ValueError('key must be != None.')
        tstamp, value_pos, value_sz = self.live_file.write(row_type, key, value)
        if value != TOMBSTONE:
            idx_value = (self.live_file.file_id, tstamp, value_sz, value_pos)
            self._keydir[(row_type, key)] = idx_value

    def get(self, row_type, key):
        """Get the value for the specified row_type, key."""
        if key is None:
            raise ValueError('key must be != None.')
        file_id, tstamp, value_sz, value_pos = self._keydir[(row_type, key)]
        value = self._get_value(file_id, value_pos, value_sz)
        return value

    def delete(self, row_type, key):
        """Delete the key/value specified by key."""
        if key is None:
            raise ValueError('key must be != None.')
        file_id, tstamp, value_sz, value_pos = self._keydir[(row_type, key)]
        self.put(row_type, key, TOMBSTONE)
        del self._keydir[(row_type, key)]


    def merge(self, immutable_files):
        """Merge a set of immutable files into a single one."""
        def by_file_id(item):
            file_id = item[1][0]
            return file_id in immutable_files
        filtered_keydir = itertools.ifilter(by_file_id, self._keydir.items())

        dest_file = TempDataFile(self.base_path)
        hint_file = dest_file.get_hint_file()
        for keydir_key, keydir_value in sorted(filtered_keydir,
                                           key=lambda item: item[1][0]):
            row_type, key = keydir_key
            _file_id, _tstamp, _value_sz, _value_pos = keydir_value
            value = self._get_value(_file_id, _value_pos, _value_sz)
            tstamp, value_pos, value_sz = dest_file.write(row_type, key, value)
            hint_entry = (key, tstamp, len(key), row_type, value_sz, value_pos)
            hint_file.write(*hint_entry)
        # close and flush files to disk
        hint_file.close()
        dest_file.close()
        new_data_file = dest_file.make_immutable()
        # once we are sure everything is on disk, mark merged files as
        # zombies, so we can still use it
        for ifile in immutable_files.values():
            ifile.make_zombie()
        return new_data_file


